from abc import ABC, abstractmethod
from argparse import ArgumentParser
from typing import Dict, Any
from pyspark.sql import SparkSession
from logging import Logger

from databricks_common.utils.logger_utils import get_logger
from databricks_common.utils.get_spark import spark

_logger = get_logger()


class MetastoreTable:  # TODO: Add tests, add shallow/deep delta clone function
    """
    Class representing a table in the Unity Catalog Metastore. This class is used to reference tables,
    and perform operations on them. It is not a Spark DataFrame, a PySpark DataFrame, or a Pandas DataFrame.
    It offers functionality such as:
    - Checking that the table exists
    - Describing the table
    - Describing the table history
    - Describing the table extended
    - Read the table as a Spark DataFrame
    """

    def __init__(self, name: str, catalog: str, schema: str) -> None:
        self.name = name
        self.catalog = catalog
        self.schema = schema
        self.ref = self._get_ref()
        self.short_ref = self._get_short_ref()

    def __repr__(self) -> str:
        cls = self.__class__.__name__
        return f"{cls}({self.ref})"

    @classmethod
    def from_string(cls, ref: str) -> "MetastoreTable":
        """
        Alternative constructor to initialise a TableReference object from a string.
        """
        cat, schema, table = ref.split(".")
        return cls(cat, schema, table)

    def check_exists(self) -> bool:
        MetastoreCatalog(self.catalog).set()
        return spark.catalog.tableExists(self.short_ref)

    def drop(self) -> None:
        try:
            spark.sql(f"DROP TABLE {self.ref}")
            print(f"NOTE: Dropped table '{self.ref}' from Unity Metastore\n-----")
        except Exception as e:
            print(
                f"NOTE: Could not drop table '{self.ref}' from Unity Metastore\n-----"
            )

    def _create_basic_table(self, comment: str = "") -> None:
        """
        Function to create a basic table at self.ref
        """
        spark.sql(
            f"""
        CREATE TABLE {self.ref}  
            (id INT, name STRING, value DOUBLE);  
        """
        )

        spark.sql(
            f"""
        INSERT INTO {self.ref}
            VALUES (1, "Yve", 1.0), 
                (2, "Omar", 2.5), 
                (3, "Elia", 3.3)
        """
        )

    def _create(self, comment: str = "") -> None:
        """
        Function to create a table in the catalog. Not intended to be used other than to create a table with the name.
        Most of the time you'll be creating tables directly in your code and won't need this function.
        """
        spark.sql(
            f"""
            CREATE TABLE {self.ref}
            USING DELTA
            COMMENT '{comment}'
            """
        )

    def describe(self) -> None:
        """
        Function to describe a table in Unity Metastore.
        """
        spark.sql(f"DESC TABLE {self.ref}").display()

    def describe_extended(self) -> None:
        """
        Function to describe a table in Unity Metastore.
        """
        spark.sql(f"DESC EXTENDED {self.ref}").display()

    def describe_history(self) -> None:
        """
        Function to describe a schema in Unity Metastore.
        """
        spark.sql(f"DESC HISTORY {self.ref}").display()

    def read(self):  # noqa
        """
        Function to read a table from the catalog.
        """
        return spark.read.table(self.ref)

    def _get_ref(self) -> str:
        return f"{self.catalog}.{self.schema}.{self.name}"

    def _get_short_ref(self) -> str:
        return f"{self.schema}.{self.name}"

    def _get_checkpoint_location(self) -> str:
        try:
            return self.checkpoint_location
        except AttributeError:
            return f"{self.catalog}.chkpts/{self.schema}/{self.name}"


class MetastoreCatalog:
    """
    Class representing a catalog in the Unity Catalog Metastore. This class is used to reference catalogs
    and perform operations on them. It offers functionality such as:
    - Checking that the catalog exists
    - Creating a catalog
    - Creating a catalog in a managed location
    - Creating a schema in a catalog
    - Describing a catalog
    - Describing a catalog in extended format
    - Dropping a catalog
    - Showing the available schemas in a catalog
    - For demo purposes, creating a catalog in a managed location
    """

    def __init__(self, name) -> None:
        """
        Parameters
        ----------
        name : str
            Name of the catalog
        """
        self.name = name

    def check_exists(self) -> bool:
        """
        Function to check if a catalog exists in Unity Metastore.
        """
        try:
            spark.sql(f"DESC CATALOG {self.name}")
            return True
        except (
            Exception
        ) as e:  # TODO: Choose specific exception in case there is a specific error
            return False

    def create_if_not_exists(self, comment: str = "") -> None:
        """
        Function to create a catalog in Unity Metastore.
        """
        if self.check_exists():
            _logger.warn(f"Catalog '{self.name}' already exists")
            return
        else:
            _logger.info(f"Creating catalog '{self.name}'")
            spark.sql(f"CREATE CATALOG {self.name} COMMENT '{comment}'")

    def create_schema(self, schema_name: str, comment: str = "") -> None:
        """
        Function to create a schema in the catalog.
        """
        self.set()
        _logger.info(f"Creating schema '{schema_name}' in catalog '{self.name}'")
        spark.sql(f"CREATE SCHEMA {schema_name} COMMENT '{comment}'")

    def create_managed_external(
        self, container_uri: str = None, container_dir: str = None, comment: str = ""
    ) -> None:
        """
        Function to create a catalog in a managed location in Unity Metastore.
        """
        if self.check_exists():
            _logger.info(f"Catalog '{self.name}' already exists")
            return
        else:
            _logger.info(
                f"Creating catalog '{self.name}' in managed location '{container_uri}/{container_dir}' with comment '{comment}'"
            )

            spark.sql(
                f"""
                CREATE CATALOG IF NOT EXISTS {self.name}
                MANAGED LOCATION '{container_uri}/{container_dir}/'
                COMMENT '{comment}' 
                """
            )

    def demo_create_managed_external_catalog(self) -> None:  # TODO: Convert to test
        """
        Function for demo purposes.
        """
        self.create_managed_external(
            container_uri="abfss://xavierarmitage-container@oneenvadls.dfs.core.windows.net",
            container_dir="lake23",
            comment="Demo comment",
        )

    def describe(self) -> None:
        """
        Function to describe a catalog in Unity Metastore.
        """
        spark.sql(f"DESC CATALOG {self.name}").display()

    def describe_extended(self) -> None:
        """
        Function to describe a catalog in Unity Metastore.
        """
        spark.sql(f"DESC CATALOG EXTENDED {self.name}").display()

    def drop(self) -> None:
        """
        Function to drop a catalog in Unity Catalog.
        """
        _logger.info(f"Dropping catalog '{self.name}'")
        spark.sql(f"DROP CATALOG IF EXISTS {self.name}")

    def drop_cascade(self) -> None:
        """
        Function to drop a catalog in Unity Catalog.
        """
        _logger.info(f"Dropping catalog '{self.name}'")
        self.drop(cascade=True)

    def drop_schema(self, cascade: bool = False):
        """
        Function to drop a schema in the catalog.
        """
        self.set()
        _logger.info(f"Dropping schema '{self.name}'")
        if cascade:
            spark.sql(f"DROP SCHEMA IF EXISTS {self.name} CASCADE")
        else:
            spark.sql(f"DROP SCHEMA IF EXISTS {self.name}")

    def set(self) -> None:
        """
        Function to set the current catalog.
        """
        spark.sql(f"SET CATALOG {self.name}")

    def show_schemas(self) -> None:
        """
        Function to show schemas in catalog.
        """
        self.set()
        spark.sql("SHOW SCHEMAS").display()


class MetastoreSchema:
    """
    Class to represent a schema in Unity Metastore.
    """

    def __init__(self, catalog: str | MetastoreCatalog, name: str) -> None:
        if isinstance(catalog, MetastoreCatalog):
            self.catalog = catalog.name
        else:
            self.catalog = catalog
        self.name = name
        self.ref = self._get_ref()

    def _set_catalog(self) -> None:
        """
        Function to set the current catalog.
        """
        MetastoreCatalog(self.catalog).set()

    def check_exists(self) -> bool:
        """
        Function to check if a schema exists in Unity Metastore.
        """
        # TODO: There might be a more direct way to do this
        self._set_catalog()
        try:
            spark.sql(f"DESC SCHEMA {self.name}")
            return True
        except Exception as e:
            return False

    def create_if_not_exists(self, comment: str = "") -> None:
        """
        Function to create a schema in Unity Metastore.
        """
        self._set_catalog()
        if self.check_exists():
            _logger.warn(f"Schema '{self.name}' already exists")
            return
        else:
            _logger.info(f"Creating schema '{self.name}'")
            spark.sql(f"CREATE SCHEMA {self.name} COMMENT '{comment}'")

    def describe(self) -> None:
        """
        Function to describe a schema in Unity Metastore.
        """
        self._set_catalog()
        spark.sql(f"DESC SCHEMA {self.name}").display()

    def describe_extended(self) -> None:
        """
        Function to describe a schema in Unity Metastore.
        """
        self._set_catalog()
        spark.sql(f"DESC SCHEMA EXTENDED {self.name}").display()

    def drop(self, cascade: bool = False) -> None:
        """
        Function to drop a schema in Unity Metastore.
        """
        self._set_catalog()
        _logger.info(f"Dropping schema '{self.name}'")
        if cascade:
            spark.sql(f"DROP SCHEMA IF EXISTS {self.name} CASCADE")
        else:
            spark.sql(f"DROP SCHEMA IF EXISTS {self.name}")

    def show_tables(self) -> None:
        """
        Function to show tables in schema.
        """
        self._set_catalog()
        spark.sql(f"SHOW TABLES IN {self.name}").display()

    def set(self) -> None:
        """
        Function to set the current schema.
        """
        self._set_catalog()
        _logger.info(f"Setting schema '{self.name}' as current schema")
        spark.sql(f"SET SCHEMA {self.name}")

    def _get_ref(self) -> str:
        return f"{self.catalog}.{self.name}"
